import { Request, Response } from "express";
import Joi, { date } from "joi";
import { Choice, Location, Question, Team, User, addUser as addUserDb } from "../db";
import { pagination } from "../utils/pagination";
import { AbstractDataType } from "sequelize";

function showDate(date: any): string {
  return (date as Date).toLocaleString('fr-FR');
}

/**
 * Handles the GET /admin path.
 * 
 * @param req The Express request object.
 * @param res The Express response object.
 */
export async function showAdminPage(req: Request, res: Response) {
  res.render("admin/index");
}

/**
 * Handles the GET /admin/teams path.
 * 
 * @param req The Express request object.
 * @param res The Express response object.
 */
export async function showAllTeams(req: Request, res: Response) {
  const teamsPerPage = 5;
  const teamCount = await Team.count();
  const maxPageCount = Math.ceil(teamCount / teamsPerPage);

  const page = Math.min(maxPageCount, Math.max(1, Number.parseInt((req.query.p || "1") as string)));

  const teams = (await Team.findAll({
    order: [['id', 'ASC']],
    offset: (page - 1) * teamsPerPage,
    limit: teamsPerPage
  })).map(team => {
    const frontendTeam = {
      id: team.id,
      name: team.id
    };
    return frontendTeam;
  })

  res.render("admin/teams", { teams, pagination: pagination(page, teamCount) });
}

const TeamParamsSchema = Joi.object({
  id: Joi.string().required()
});

/**
 * Handles the GET /admin/team/show/:id path.
 * 
 * @param req The Express request object.
 * @param res The Express response object.
 */
export async function showTeam(req: Request, res: Response) {
  try {
    const { id } = await TeamParamsSchema.validateAsync(req.params);
    const team = await Team.findOne({ where: { id } });
    if (team == null)
      throw Error("invalid team ID");

    const frontendTeam = {
      id: team.id,
      name: team.id,
      players: (await team.getPlayers()).map(player => {
        const frontendUser = {
          id: player.id,
          login: player.id,
          admin: player.admin
        };
        return frontendUser;
      }
      )
    };

    res.render("admin/team", { team: frontendTeam });
  } catch (err) {
    res.status(400).render("internalError", { msg: err });
    return;
  }
}

/**
 * Handles the GET /admin/team/delete/:id path.
 * 
 * @param req The Express request object.
 * @param res The Express response object.
 */
export async function deleteTeam(req: Request, res: Response) {
  try {
    const { id } = await TeamParamsSchema.validateAsync(req.params);
    const team = await Team.findOne({ where: { id } });
    if (team == null)
      throw Error("invalid team ID");

    team.destroy();
    res.redirect("/admin");
  } catch (err) {
    res.status(400).render("internalError", { msg: err });
    return;
  }
}

/**
 * Handles the GET /admin/team/add path.
 * 
 * @param req The Express request object.
 * @param res The Express response object.
 */
export async function showAddTeamForm(req: Request, res: Response) {
  res.render("admin/addTeam")
}

const AddTeamFormBodySchema = Joi.object({
  name: Joi.string().required()
});

/**
 * Handles the POST /admin/team/add path.
 * 
 * @param req The Express request object.
 * @param res The Express response object.
 */
export async function addTeam(req: Request, res: Response) {
  try {
    const { name } = await AddTeamFormBodySchema.validateAsync(req.body);
    const team = await Team.create({ id: name });
    res.redirect('/admin');
  } catch (err) {
    res.status(400).render("internalError", { msg: err });
    return;
  }
}

/**
 * Handles the GET /admin/users path.
 * 
 * @param req The Express request object.
 * @param res The Express response object.
 */
export async function showAllUsers(req: Request, res: Response) {
  const usersPerPage = 5;
  const userCount = await User.count();
  const maxPageCount = Math.ceil(userCount / usersPerPage);

  const page = Math.min(maxPageCount, Math.max(1, Number.parseInt((req.query.p || "1") as string)));

  const users = (await User.findAll({
    order: [['id', 'ASC']],
    offset: (page - 1) * usersPerPage,
    limit: usersPerPage
  })).map(user => {
    const frontendUser = {
      id: user.id,
      login: user.id,
      admin: user.admin
    };
    return frontendUser;
  })

  res.render("admin/users", { users, pagination: pagination(page, userCount) });
}

const UserParamsSchema = Joi.object({
  id: Joi.string().required()
});

/**
 * Handles the GET /admin/user/show/:id path.
 * 
 * @param req The Express request object.
 * @param res The Express response object.
 */
export async function showUser(req: Request, res: Response) {
  try {
    const { id } = await UserParamsSchema.validateAsync(req.params);
    const user = await User.findOne({ where: { id } });
    if (user == null)
      throw Error("invalid user ID");

    const userTeam = await user.getTeam();
    let frontendTeam = undefined;
    if (userTeam) {
      frontendTeam = {
        id: userTeam.id,
        name: userTeam.id,
      };
    }

    const frontendUser = {
      id: user.id,
      login: user.id,
      admin: user.admin,
      team: frontendTeam
    };

    res.render("admin/user", { user: frontendUser });
  } catch (err) {
    res.status(400).render("internalError", { msg: err });
    return;
  }
}

/**
 * Handles the GET /admin/user/delete/:id path.
 * 
 * @param req The Express request object.
 * @param res The Express response object.
 */
export async function deleteUser(req: Request, res: Response) {
  try {
    const { id } = await UserParamsSchema.validateAsync(req.params);
    const user = await User.findOne({ where: { id } });
    if (user == null)
      throw Error("invalid user ID");

    user.destroy();
    res.redirect("/admin");
  } catch (err) {
    res.status(400).render("internalError", { msg: err });
    return;
  }
}

/**
 * Handles the GET /admin/user/add path.
 * 
 * @param req The Express request object.
 * @param res The Express response object.
 */
export async function showAddUserForm(req: Request, res: Response) {
  const teams = (await Team.findAll()).map(team => {
    const frontendTeam = {
      id: team.id,
      name: team.id
    };
    return frontendTeam;
  });

  res.render("admin/addUser", { teams });
}

const AddUserFormBodySchema = Joi.object({
  login: Joi.string().required(),
  password: Joi.string().required(),
  team: Joi.string().allow("none").required(),
  isAdmin: Joi.boolean().default(false)
});

/**
 * Handles the POST /admin/user/add path.
 * 
 * @param req The Express request object.
 * @param res The Express response object.
 */
export async function addUser(req: Request, res: Response) {
  try {
    const { login, password, team, isAdmin } = await AddUserFormBodySchema.validateAsync(req.body);
    let user = await addUserDb(login, password, isAdmin);

    if (team !== "none") {
      user.setTeam(team);
      user.save();
    }

    res.redirect('/admin');
  } catch (err) {
    res.status(400).render("internalError", { msg: err });
    return;
  }
}

/**
 * Handles the POST /admin/locations.
 * 
 * @param req The Express request object.
 * @param res The Express response object.
 */
export async function showAllLocations(req: Request, res: Response) {
  const locationsPerPage = 5;
  const locationCount = await Location.count();
  const maxPageCount = Math.ceil(locationCount / locationsPerPage);

  const page = Math.min(maxPageCount, Math.max(1, Number.parseInt((req.query.p || "1") as string)));

  const locations = (await Location.findAll({
    order: [['userName', 'ASC']],
    offset: (page - 1) * locationsPerPage,
    limit: locationsPerPage
  })).map(location => {
    const frontendTeam = {
      id: location.id,
      name: location.userName
    };
    return frontendTeam;
  })

  res.render("admin/locations", { locations, pagination: pagination(page, locationCount) });
}

const LocationParamsSchema = Joi.object({
  id: Joi.string().required()
});

/**
 * Handles the POST /admin/location/show/:id path.
 * 
 * @param req The Express request object.
 * @param res The Express response object.
 */
export async function showLocation(req: Request, res: Response) {
  try {
    const { id } = await LocationParamsSchema.validateAsync(req.params);
    const location = await Location.findOne({ where: { id } });
    if (location == null)
      throw Error("invalid location ID");

    const frontendLocation = {
      id: location.id,
      name: location.userName,
      questions: (await location.getQuestions()).map(question => {
        const frontendQuestion = {
          id: question.id,
          text: question.text,
          dateStart: showDate(question.validDateStart),
          dateEnd: showDate(question.validDateEnd),
          allowMultipleChoices: question.allowMultipleChoices
        };
        return frontendQuestion;
      })
    };

    res.render("admin/location", { location: frontendLocation });
  } catch (err) {
    res.status(400).render("internalError", { msg: err });
    return;
  }
}

/**
 * Handles the POST /admin/location/delete/:id path.
 * 
 * @param req The Express request object.
 * @param res The Express response object.
 */
export async function deleteLocation(req: Request, res: Response) {
  try {
    const { id } = await LocationParamsSchema.validateAsync(req.params);
    const location = await Location.findOne({ where: { id } });
    if (location == null)
      throw Error("invalid location ID");

    location.destroy();
    res.redirect("/admin");
  } catch (err) {
    res.status(400).render("internalError", { msg: err });
    return;
  }
}

/**
 * Handles the GET /admin/location/add path.
 * 
 * @param req The Express request object.
 * @param res The Express response object.
 */
export async function showAddLocationForm(req: Request, res: Response) {
  res.render("admin/addLocation");
}

const AddLocationFormBodySchema = Joi.object({
  name: Joi.string().required()
});

/**
 * Handles the POST /admin/location/add path.
 * 
 * @param req The Express request object.
 * @param res The Express response object.
 */
export async function addLocation(req: Request, res: Response) {
  try {
    const { name } = await AddLocationFormBodySchema.validateAsync(req.body);
    const location = await Location.create({ userName: name });
    res.redirect('/admin');
  } catch (err) {
    res.status(400).render("internalError", { msg: err });
    return;
  }
}

const QuestionParamsSchema = Joi.object({
  id: Joi.string().required()
});

/**
 * Handles the GET /admin/question/show/:id path.
 * 
 * @param req The Express request object.
 * @param res The Express response object.
 */
export async function showQuestion(req: Request, res: Response) {
  try {
    const { id } = await QuestionParamsSchema.validateAsync(req.params);
    const question = await Question.findOne({ where: { id } });
    if (question == null)
      throw Error("invalid question ID");

    const questionLocation = await question.getLocation();
    const frontendLocation = {
      id: questionLocation.id,
      name: questionLocation.userName,
    };

    const frontendQuestion = {
      id: question.id,
      text: question.text,
      allowMultipleChoices: question.allowMultipleChoices,
      dateStart: showDate(question.validDateStart),
      dateEnd: showDate(question.validDateEnd),
      location: frontendLocation,
      choices: (await question.getChoices()).map(choice => {
        const frontendChoice = {
          id: choice.id,
          text: choice.text,
          isAnswer: choice.isAnswer
        };
        return frontendChoice;
      })
    };

    res.render("admin/question", { question: frontendQuestion });
  } catch (err) {
    res.status(400).render("internalError", { msg: err });
    return;
  }
}

/**
 * Handles the GET /admin/question/delete/:id path.
 * 
 * @param req The Express request object.
 * @param res The Express response object.
 */
export async function deleteQuestion(req: Request, res: Response) {
  try {
    const { id } = await QuestionParamsSchema.validateAsync(req.params);
    const question = await Question.findOne({ where: { id } });
    if (question == null)
      throw Error("invalid question ID");

    question.destroy();
    res.redirect("/admin");
  } catch (err) {
    res.status(400).render("internalError", { msg: err });
    return;
  }
}

/**
 * Handles the GET /admin/question/add path.
 * 
 * @param req The Express request object.
 * @param res The Express response object.
 */
export async function showAddQuestionForm(req: Request, res: Response) {
  res.render("admin/addQuestion", {
    locations: await Location.findAll()
  });
}

const AddQuestionFormBodySchema = Joi.object({
  location: Joi.string().required(),
  text: Joi.string().required(),
  dateBegin: Joi.date().required(),
  dateEnd: Joi.date().required(),
  isQCM: Joi.boolean().default(false)
});

/**
 * Handles the POST /admin/question/add path.
 * 
 * @param req The Express request object.
 * @param res The Express response object.
 */
export async function addQuestion(req: Request, res: Response) {
  try {
    const { location, text, dateBegin, dateEnd, isQCM } = await AddQuestionFormBodySchema.validateAsync(req.body);
    const question = await Question.create({ locationId: location, text, validDateStart: dateBegin, validDateEnd: dateEnd, allowMultipleChoices: isQCM });
    res.redirect(`/admin/question/show/${question.id}`);
  } catch (err) {
    res.status(400).render("internalError", { msg: err });
    return;
  }
}

/**
 * Handles the GET /admin/choice/add/:id path.
 * 
 * @param req The Express request object.
 * @param res The Express response object.
 */
export async function showAddChoiceForm(req: Request, res: Response) {
  try {
    const id = req.params.id;
    const question = await Question.findOne({ where: { id } });
    if (question == null)
      throw Error("invalid question ID");

    const frontendQuestion = {
      id: question.id,
      text: question.text
    };

    res.render("admin/addChoice", { question: frontendQuestion });
  } catch (err) {
    res.status(400).render("internalError", { msg: err });
    return;
  }
}

const AddChoiceFormBodySchema = Joi.object({
  text: Joi.string().required(),
  isAnswer: Joi.boolean().default(false)
});

/**
 * Handles the POST /admin/choice/add/:id path.
 * 
 * @param req The Express request object.
 * @param res The Express response object.
 */
export async function addChoice(req: Request, res: Response) {
  try {
    const questionId = req.params.id;
    const { text, isAnswer } = await AddChoiceFormBodySchema.validateAsync(req.body);
    const choice = await Choice.create({ questionId: ((questionId as unknown) as AbstractDataType), text, isAnswer });
    res.redirect(`/admin/question/show/${questionId}`);
  } catch (err) {
    res.status(400).render("internalError", { msg: err });
    return;
  }
}
