import { Request, Response } from "express";
import Joi from "joi";
import { loginUser } from "../db";

/**
 * Handles the GET /login path.
 * 
 * @param req The Express request object.
 * @param res The Express response object.
 */
export async function renderLogin(req: Request, res: Response) {
  const redirect = req.query.r;
  if (redirect) {
    if (!req.session)
      req.session = {};
    req.session.redirect = redirect;
  }

  res.render("login");
}

const LoginFormBodySchema = Joi.object({
  login: Joi.string().required(),
  password: Joi.string().required()
});

/**
 * Handles the POST /login path.
 * 
 * @param req The Express request object.
 * @param res The Express response object.
 */
export async function login(req: Request, res: Response) {
  try {
    const { login, password } = await LoginFormBodySchema.validateAsync(req.body);
    const user = await loginUser(login, password);
    if (user == null)
      throw Error("invalid login or password");

    if (!req.session)
      req.session = {};
    req.session.user = {
      id: user.id,
      login: user.id,
      admin: user.admin
    };

    const redirect = req.session.redirect || '/';
    req.session.redirect = undefined; // clear redirect info

    res.redirect(redirect);
  } catch (err) {
    res.status(401)
      .render("login", { authFailed: true });
  }
}

/**
 * Handles the GET /logout path.
 * 
 * @param req The Express request object.
 * @param res The Express response object.
 */
export async function logout(req: Request, res: Response) {
  // Clear user session data.
  if (req.session)
    req.session.user = undefined;

  res.redirect('/');
}
