import crypto from "crypto";
import { AbstractDataType, Association, BelongsToCreateAssociationMixin, BelongsToGetAssociationMixin, BelongsToSetAssociationMixin, CreationOptional, DataTypes, ForeignKey, HasManyAddAssociationMixin, HasManyAddAssociationsMixin, HasManyCountAssociationsMixin, HasManyCreateAssociationMixin, HasManyGetAssociationsMixin, HasManyHasAssociationMixin, HasManyHasAssociationsMixin, HasManyRemoveAssociationMixin, HasManyRemoveAssociationsMixin, HasManySetAssociationsMixin, InferAttributes, InferCreationAttributes, Model, NonAttribute, Sequelize, UUIDV4 } from "sequelize";
import { promisify } from "util";
import process from "process";

const scrypt = promisify(crypto.scrypt) as unknown as ((password: crypto.BinaryLike, salt: crypto.BinaryLike, keylen: number) => Promise<Buffer>);

const sequelize = new Sequelize({
  dialect: "sqlite",
  storage: process.env.DB_FILE || "kahulm.sqlite",
  transactionType: ('IMMEDIATE' as any)
});

// const sequelize = new Sequelize({
//   dialect: "sqlite",
//   storage: ":memory:"
// });

export class User extends Model<InferAttributes<User>, InferCreationAttributes<User>> {
  declare id: string;
  declare passwordHash: string;
  declare passwordSalt: string;
  declare admin: CreationOptional<boolean>;
  declare teamId: CreationOptional<ForeignKey<Team["id"]>>;

  declare static associations: {
    team: Association<User, Team>;
  };

  declare team?: NonAttribute<Team>;
  declare getTeam: BelongsToGetAssociationMixin<Team>;
  declare setTeam: BelongsToSetAssociationMixin<Team, number>;
  declare createTeam: BelongsToCreateAssociationMixin<Team>;
}

export class Location extends Model<InferAttributes<Location>, InferCreationAttributes<Location>> {
  declare id: CreationOptional<AbstractDataType>;
  declare userName: string;

  declare static associations: {
    questions: Association<Location, Question>;
  };

  declare questions?: NonAttribute<Question[]>;
  declare getQuestions: HasManyGetAssociationsMixin<Question>;
  declare addQuestion: HasManyAddAssociationMixin<Question, never>;
  declare addQuestions: HasManyAddAssociationsMixin<Question, never>;
  declare setQuestions: HasManySetAssociationsMixin<Question, never>;
  declare removeQuestion: HasManyRemoveAssociationMixin<Question, never>;
  declare removeQuestions: HasManyRemoveAssociationsMixin<Question, never>;
  declare hasQuestion: HasManyHasAssociationMixin<Question, never>;
  declare hasQuestions: HasManyHasAssociationsMixin<Question, never>;
  declare countQuestions: HasManyCountAssociationsMixin;
  declare createQuestion: HasManyCreateAssociationMixin<Question, "locationId">;
}

export class Question extends Model<InferAttributes<Question>, InferCreationAttributes<Question>> {
  declare id: CreationOptional<AbstractDataType>;
  declare imageIdx?: number;
  declare locationId: ForeignKey<Location['id']>;
  declare validDateStart: DataTypes.DateDataType;
  declare validDateEnd: DataTypes.DateDataType;
  declare allowMultipleChoices: CreationOptional<boolean>;
  declare text: string;

  declare static associations: {
    location: Association<Question, Location>;
    choices: Association<Question, Choice>;
  };

  declare location?: NonAttribute<Location>;
  declare getLocation: BelongsToGetAssociationMixin<Location>;
  declare setLocation: BelongsToSetAssociationMixin<Location, DataTypes.AbstractDataType>;
  declare createLocation: BelongsToCreateAssociationMixin<Location>;

  declare choices?: NonAttribute<Choice[]>;
  declare getChoices: HasManyGetAssociationsMixin<Choice>;
  declare addChoice: HasManyAddAssociationMixin<Choice, never>;
  declare addChoices: HasManyAddAssociationsMixin<Choice, never>;
  declare setChoices: HasManySetAssociationsMixin<Choice, never>;
  declare removeChoice: HasManyRemoveAssociationMixin<Choice, never>;
  declare removeChoices: HasManyRemoveAssociationsMixin<Choice, never>;
  declare hasChoice: HasManyHasAssociationMixin<Choice, never>;
  declare hasChoices: HasManyHasAssociationsMixin<Choice, never>;
  declare countChoices: HasManyCountAssociationsMixin;
  declare createChoice: HasManyCreateAssociationMixin<Choice, "questionId">;
}

export class Choice extends Model<InferAttributes<Choice>, InferCreationAttributes<Choice>> {
  declare id: CreationOptional<AbstractDataType>;
  declare questionId: ForeignKey<Question['id']>;
  declare isAnswer: CreationOptional<boolean>;
  declare text: string;

  declare static associations: {
    question: Association<Choice, Question>;
  };

  declare question?: NonAttribute<Question>;
  declare getQuestion: BelongsToGetAssociationMixin<Question>;
  declare setQuestion: BelongsToSetAssociationMixin<Question, DataTypes.AbstractDataType>;
  declare createQuestion: BelongsToCreateAssociationMixin<Question>;
}

export class Answer extends Model<InferAttributes<Answer>, InferCreationAttributes<Answer>> {
  declare playerId: ForeignKey<User['id']>;
  declare questionId: ForeignKey<Question['id']>;
  declare selectedChoices: string; // stringified JSON list of choice IDs, e.g. ['first-id', 'second-id', ...]
  declare isCorrect: boolean;

  declare static associations: {
    player: Association<Answer, User>;
    question: Association<Answer, Question>;
  };

  declare player?: NonAttribute<User>;
  declare getPlayer: BelongsToGetAssociationMixin<User>;
  declare setPlayer: BelongsToSetAssociationMixin<User, number>;
  declare createPlayer: BelongsToCreateAssociationMixin<User>;

  declare question?: NonAttribute<Question>;
  declare getQuestion: BelongsToGetAssociationMixin<Question>;
  declare setQuestion: BelongsToSetAssociationMixin<Question, DataTypes.AbstractDataType>;
  declare createQuestion: BelongsToCreateAssociationMixin<Question>;
}

export class Team extends Model<InferAttributes<Team>, InferCreationAttributes<Team>> {
  declare id: string;

  declare static associations: {
    players: Association<Team, User>;
  };

  declare players?: NonAttribute<User[]>;
  declare getPlayers: HasManyGetAssociationsMixin<User>;
  declare addPlayer: HasManyAddAssociationMixin<User, never>;
  declare addPlayers: HasManyAddAssociationsMixin<User, never>;
  declare setPlayers: HasManySetAssociationsMixin<User, never>;
  declare removePlayer: HasManyRemoveAssociationMixin<User, never>;
  declare removePlayers: HasManyRemoveAssociationsMixin<User, never>;
  declare hasPlayer: HasManyHasAssociationMixin<User, never>;
  declare hasPlayers: HasManyHasAssociationsMixin<User, never>;
  declare countPlayers: HasManyCountAssociationsMixin;
  declare createPlayer: HasManyCreateAssociationMixin<User, "teamId">;
}

User.init(
  {
    id: {
      type: DataTypes.STRING,
      primaryKey: true,
    },
    passwordHash: {
      type: DataTypes.STRING,
      allowNull: false,
    },
    passwordSalt: {
      type: DataTypes.STRING,
      allowNull: false,
    },
    admin: {
      type: DataTypes.BOOLEAN,
      allowNull: false,
      defaultValue: false,
    },
    teamId: {
      type: DataTypes.TEXT,
      allowNull: true,
      defaultValue: null
    }
  },
  { sequelize }
);

Location.init(
  {
    id: {
      type: DataTypes.UUID,
      defaultValue: DataTypes.UUIDV4,
      primaryKey: true,
    },

    userName: {
      type: DataTypes.STRING
    }
  },
  { sequelize }
);

Question.init(
  {
    id: {
      type: DataTypes.UUID,
      defaultValue: DataTypes.UUIDV4,
      primaryKey: true,
    },
    imageIdx: {
      type: DataTypes.INTEGER,
      allowNull: true,
      defaultValue: null
    },
    locationId: {
      type: DataTypes.UUID,
      allowNull: false
    },
    validDateStart: {
      type: DataTypes.DATE,
      defaultValue: () => Date.now(),
      allowNull: false
    },
    validDateEnd: {
      type: DataTypes.DATE,
      defaultValue: () => Date.now(),
      allowNull: false
    },
    allowMultipleChoices: {
      type: DataTypes.BOOLEAN,
      defaultValue: false,
      allowNull: false
    },
    text: {
      type: DataTypes.TEXT,
      allowNull: false,
    }
  },
  { sequelize }
);

Choice.init(
  {
    id: {
      type: DataTypes.UUID,
      defaultValue: DataTypes.UUIDV4,
      primaryKey: true,
    },
    questionId: {
      type: Question.getAttributes().id.type,
      onDelete: "CASCADE",
      allowNull: false
    },
    isAnswer: {
      type: DataTypes.BOOLEAN,
      defaultValue: false,
      allowNull: false
    },
    text: {
      type: DataTypes.TEXT,
      allowNull: false,
    }
  },
  { sequelize }
);

Answer.init(
  {
    playerId: {
      type: User.getAttributes().id.type,
      primaryKey: true
    },
    questionId: {
      type: Question.getAttributes().id.type,
      primaryKey: true
    },
    selectedChoices: {
      type: DataTypes.STRING,
      allowNull: false
    },
    isCorrect: {
      type: DataTypes.BOOLEAN,
      allowNull: false
    }
  },
  { sequelize }
)

Team.init(
  {
    id: {
      type: DataTypes.TEXT,
      primaryKey: true,
    }
  },
  { sequelize }
);

Team.hasMany(User, {
  foreignKey: "teamId",
  as: "players"
});

User.belongsTo(Team, {
  foreignKey: "teamId",
  as: "team"
});

Location.hasMany(Question, {
  foreignKey: "locationId",
  as: "questions"
});

Question.belongsTo(Location, {
  foreignKey: "locationId",
  as: "location"
});

Question.hasMany(Choice, {
  foreignKey: "questionId",
  as: "choices",
});

Choice.belongsTo(Question, {
  foreignKey: "questionId",
  as: "question",
});

Answer.belongsTo(Question, {
  foreignKey: "questionId",
  as: "question",
});

Answer.belongsTo(User, {
  foreignKey: "playerId",
  as: "player",
});

export async function initDatabase() {
  // FIXME not for production :)
  await sequelize.sync();

  // Add admin user if requested.
  if (process.env.ADMIN_USERNAME && process.env.ADMIN_PASSWORD) {
    console.log("Admin user created using login ", process.env.ADMIN_USERNAME);
    await addUser(process.env.ADMIN_USERNAME, process.env.ADMIN_PASSWORD, true);
  }
}

function parseOptionalInt(value: any): number | undefined {
  if (value) {
    return parseInt(value);
  } else {
    return undefined;
  }
}

const saltLength = parseOptionalInt(process.env.PASSWORD_SALT_LENGTH) || 128;
const passwordHashLength = parseOptionalInt(process.env.PASSWORD_HASH_LENGTH) || 64;

function normalizeLogin(login: string): string {
  return login.trim().toLowerCase().normalize();
}

export async function addUser(login: string, password: string, admin: boolean = false): Promise<User> {
  const salt = crypto.randomBytes(saltLength);
  const hash = await scrypt(password.normalize(), salt, passwordHashLength);
  return await User.create({
    id: normalizeLogin(login),
    admin,
    passwordHash: hash.toString("base64"),
    passwordSalt: salt.toString("base64")
  });
}

export async function loginUser(login: string, password: string): Promise<User | null> {
  const user = await User.findOne({ where: { id: normalizeLogin(login) } });
  if (user === null) return null;
  const hash = Buffer.from(user.passwordHash, "base64");
  const salt = Buffer.from(user.passwordSalt, "base64");
  const otherHash = await scrypt(password.normalize(), salt, passwordHashLength);
  if (!hash.equals(otherHash)) return null;
  return user;
}

// An User object extended with an additional score field. 
// This is used by the findBestPlayers() function.
export class UserWithScore extends User {
  declare score: number;
}

export async function findBestPlayers(offset: number = 0, limit: number = 5): Promise<UserWithScore[]> {
  return (await User.findAll(
    {
      attributes: {
        include: [
          [ 
            sequelize.literal(`(
              SELECT COUNT(*)
                    FROM answers AS answer
                    WHERE
                        answer.playerId = user.id
                        AND
                        answer.isCorrect
              )`), 
            'score'
          ],
        ], 
      },
      order: [[sequelize.literal('score'), 'DESC']],
      offset,
      limit
    } 
  )).map((user) => {
    // Add the score field to the returned objects.
    let userWithScore = user as UserWithScore;
    // The following is not an error as the 'score' attribute exist. 
    // It is added by the custom SQL code.
    // @ts-expect-error
    userWithScore.score = user.getDataValue('score');
    return userWithScore;
  });
}

export class UserWithTeam extends User {
  declare name: string
}

export async function listPlayers(offset: number = 0, limit: number = 5): Promise<UserWithTeam[]> {
  return (await User.findAll(
    {
      attributes: {
        include: [
          [ 
            sequelize.literal(`(
              SELECT teams.name
                    FROM teams
                    WHERE
                        teams.id = user.teamId
              )`), 
            'name'
          ],
        ], 
      },
      order: [[sequelize.literal('name'), 'DESC']],
      offset,
      limit
    } 
  )).map((user) => {
    // Add the score field to the returned objects.
    let userWithScore = user as UserWithTeam;
    // The following is not an error as the 'score' attribute exist. 
    // It is added by the custom SQL code.
    // @ts-expect-error
    userWithScore.name = user.getDataValue("name");
    console.log("Name: %s", userWithScore.name);
    return userWithScore;
  });;
}

// A Team object extended with an additional score field. 
// This is used by the findBestTeams() function.
export class TeamWithScore extends Team {
  declare score: number;
}

export async function findBestTeams(offset: number = 0, limit: number = 5): Promise<TeamWithScore[]> {
  return (await Team.findAll(
    {
      attributes: {
        include: [
          [
            sequelize.literal(`(
              SELECT COUNT(*)
                    FROM answers JOIN users ON answers.playerId = users.id
                    WHERE
                        users.teamId = team.id
                        AND
                        answers.isCorrect
              )`), 
            'score'
          ]
        ]
      },
      order: [[sequelize.literal('score'), 'DESC']],
      offset,
      limit
    }
  )).map((team) => {
    // Add the score field to the returned objects.
    let teamWithScore = team as TeamWithScore;
    // The following is not an error as the 'score' attribute exist. 
    // It is added by the custom SQL code.
    // @ts-expect-error
    teamWithScore.score = team.getDataValue('score');
    return teamWithScore;
  });
}
