import fs from 'node:fs';
import { parse } from 'csv-parse';
import { Location, Question } from './db';
import { DateDataType } from 'sequelize';

const getLocation = async (location: string): Promise<Location> => {
  return (await Location.findOrCreate({ where: { userName: location }, defaults: { userName: location } }))[0];
}

const handleRecord = async (record: any) => {
  const DATE_REGEX = /(\d+)\/(\d+)/;
  const TIME_REGEX = /(\d+\.\d+)\s*-\s*(\d+\.\d+)/;

  const index = parseInt(record[0]);
  const text = record[1];
  const choices = [ record[2], record[3], record[4], record[5] ];
  const hasImage = record[6].length > 0;
  const locationDb = await getLocation(record[7]);

  const dateParseResult = DATE_REGEX.exec(record[8]);
  if (dateParseResult == null)
    throw Error("invalid question CSV table");
  const day = parseInt(dateParseResult[1]);
  const month = parseInt(dateParseResult[2]) - 1;
  const year = (new Date()).getFullYear();

  const timeParseResult = TIME_REGEX.exec(record[9]);
  if (timeParseResult == null)
    throw Error("invalid question CSV table");
  const hourStart = parseFloat(timeParseResult[1]);
  const hourEnd = parseFloat(timeParseResult[2]);

  const dateStart = new Date(year, month, day, hourStart);
  const dateEnd = new Date(year, month, day, hourEnd);
  
  const questionDb = await Question.create({
    text,
    imageIdx: hasImage ? index : undefined,
    locationId: locationDb.id,
    allowMultipleChoices: false,
    validDateStart: (dateStart as unknown) as DateDataType,
    validDateEnd: (dateEnd as unknown) as DateDataType,
  });

  for (let choice of choices) {
    const choiceDb = questionDb.createChoice({
      text: choice,
      isAnswer: (choice === record[2]),
    });
  }
}

export const loadQuestionTable = async (file: string) => {
  console.log("Loading questions");

  const parser = fs
    .createReadStream(file)
    .pipe(parse({ fromLine: 2 }));

  for await (const record of parser) {
    await handleRecord(record);
  }

  console.log("Questions loaded");
};

